
static ERROR RegisterProgram(struct Install *, struct XMLTag *, OBJECTPTR, STRING);

/*****************************************************************************
** Internal: TranslateMethod()
*/

static LONG TranslateMethod(STRING String)
{
   if (StrCompare("overwrite", String, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      return(METHOD_OVERWRITE);
   }
   else if (StrCompare("backup", String, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      return(METHOD_BACKUP);
   }
   else if (StrCompare("cautious", String, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      return(METHOD_CAUTIOUS);
   }
   else return(METHOD_OVERWRITE);
}

/*****************************************************************************
** Internal: CommandComponent()
*/

static ERROR CommandComponent(struct Install *Self, OBJECTPTR Output, struct XMLTag *XMLTag)
{
   struct XMLTag *tag;
   STRING name;
   LONG i;
   WORD reverse;
   ERROR error;

   DPrintF("~Installer:","Processing <component> tag.");

   name = NULL;
   reverse = FALSE;
   for (i=1; i < XMLTag->TotalAttributes; i++) {
      if (StrCompare("name", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         name = XMLTag->Attributes[i].Value;
         break;
      }
      else if (StrCompare("!name", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         name = XMLTag->Attributes[i].Value;
         reverse = TRUE;
         break;
      }
   }

   if (name) {
      for (i=0; i < Self->prvTotalVariables; i++) {
         if (StrCompare(name, Self->prvVariables[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            error = StrCompare("1", Self->prvVariables[i].Value, 0, STR_MATCHLENGTH);

            if (((error IS ERR_Okay) AND (reverse IS FALSE)) OR ((error != ERR_Okay) AND (reverse IS TRUE))) {
               /*** Execute the commands within the <component> tag ***/

               DPrintF("Installer:","Criteria passed for component selection (%s = %s).", name, Self->prvVariables[i].Value);

               for (tag=XMLTag->Child; tag; tag=tag->Next) {
                  for (i=0; glInstallCommands[i].Tag; i++) {
                     if (StrCompare(glInstallCommands[i].Tag, tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                        if ((error = glInstallCommands[i].Routine(Self, Output, tag)) != ERR_Okay) {
                           StepBack();
                           return(error);
                        }
                        break;
                     }
                  }
               }
            }
            StepBack();
            return(ERR_Okay);
         }
      }

      DPrintF("Installer:","Criteria failed for component selection \"%s\".", name);
   }

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************
** Internal: CommandCopy()
*/

static UBYTE glForceCopy;

static ERROR CommandCopy(struct Install *Self, OBJECTPTR Output, struct XMLTag *XMLTag)
{
   UBYTE buffer[260];
   struct mtDecompressFile decompress;
   struct acWrite write;
   OBJECTPTR file;
   STRING src, dest;
   LONG i, j, method, len, type, permissions;
   UBYTE deletedir, execute;
   ERROR error;
   BYTE nofail;

   DPrintF("~Installer:","Processing <copy> tag.");

   glForceCopy = FALSE;
   src    = NULL;
   dest   = NULL;
   method = METHOD_OVERWRITE;
   nofail = FALSE;
   execute = FALSE;
   permissions = PERMIT_EVERYONE_READ|PERMIT_EVERYONE_WRITE;
   for (i=1; i < XMLTag->TotalAttributes; i++) {
      if (StrCompare("src", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         src = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("dest", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         dest = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("force", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         glForceCopy = TRUE;
      }
      else if (StrCompare("method", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         method = TranslateMethod(XMLTag->Attributes[i].Value);
      }
      else if (StrCompare("nofail", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         nofail = TRUE;
      }
      else if (StrCompare("executable", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         permissions |= PERMIT_EVERYONE_EXEC;
         execute = TRUE;
      }
   }

   error = ERR_Okay;
   if ((src) AND (dest) AND (*src) AND (*dest)) {
      /* Determine if the source is from the compressed installation file,
      ** or if it is from a specific location on the user's system.
      */

      for (i=0; (src[i]) AND (src[i] != ':'); i++);

      if (src[i] != ':') {
         /* If the copy operation is to a directory and if that directory
         ** does not exist, we mark the directory for deletion during uninstall.
         */

         deletedir = FALSE;
         if (Self->prvUninstall) {
            for (i=0; dest[i]; i++);
            if ((dest[i-1] IS '/') AND (AnalyseLocation(dest, &type) != ERR_Okay)) deletedir = TRUE;
         }

         /* 'Copy' the file by decompressing it.  Note that the source file
         ** reference may use wildcards, as supported by the DecompressFile()
         ** method.
         */

         SetField(Self->prvCompression, FID_Feedback, FT_POINTER, &CopyFeedback);

            decompress.Path        = src;
            decompress.Destination = dest;
            decompress.Flags       = NULL;
            error = Action(MT_DecompressFile, Self->prvCompression, &decompress);
            if (error) DPrintF("@Install:","Error %d decompressing %s", error, src);

         SetField(Self->prvCompression, FID_Feedback, FT_POINTER, NULL);

         if (deletedir IS TRUE) {
            write.Buffer = buffer;
            write.Length = StrFormat(buffer, sizeof(buffer), "  <delete src=\"%s\" empty/>\n", dest);
            Action(AC_Write, Self->prvUninstall, &write);
         }

         /*** Set execute permissions on the file ***/

         if (execute) {
            i = StrCopy(dest, buffer, sizeof(buffer));
            if (buffer[i-1] IS '/') {
               for (j=0; src[j]; j++);
               while ((j > 0) AND (src[j-1] != '/') AND (src[j-1] != ':')) j--;
               while ((src[j]) AND (i < sizeof(buffer)-1)) buffer[i++] = src[j++];
               buffer[i] = 0;
            }

            DPrintF("Install:","Setting execute permissions on \"%s\"", buffer);

            if (CreateObject(ID_FILE, NF_CHILD, &file, NULL,
               FID_Location|TSTRING, buffer,
               TAGEND) IS ERR_Okay) {

               if (GetField(file, FID_Permissions, FT_LONG, &permissions) IS ERR_Okay) {
                  /* Note that the USERID bit is set so that the installed file will
                  ** run with the same permissions as the logged in user.  This means
                  ** that if the user logs in as root and installs program files, they
                  ** will be able to run with fast video access (due to 'suid root'
                  ** security access to /dev/mem).
                  */
                  permissions |= PERMIT_EVERYONE_EXEC;

                  #ifdef __linux__
                     if (AnalyseLocation("/dev/pcmem", NULL) != ERR_Okay) permissions |= PERMIT_USERID;
                  #endif

                  if (SetField(file, FID_Permissions, FT_LONG, &permissions) != ERR_Okay) {
                     DPrintF("@Install:","Failed to set file permissions.");
                  }
               }
               else DPrintF("@Install:","Failed to read file permissions.");

               Action(AC_Free, file, NULL);
            }
            else DPrintF("@Install:","Failed to open file \"%s\".", buffer);
         }
      }
      else {
         /*** Copy a file somewhere in the system to a new location ***/

         len = StrFormat(buffer, sizeof(buffer), "Copying \"%s\" to \"%s\".", src, dest);
         ActionTags(AC_DataChannel, Output, Self->Head.UniqueID, DATA_TEXT, 1, buffer, len + 1, 1);

         if ((error = CreateObject(ID_FILE, NF_CHILD, &file, NULL,
               FID_Location|TSTRING,  src,
               FID_Flags|TLONG,       FL_READ,
               TAGEND)) IS ERR_Okay) {
            error = ActionTags(MT_CopyFile, file, dest);
            Action(AC_Free, file, NULL);
         }
      }
   }

   StepBack();
   if (nofail) return(ERR_Okay);
   else return(error);
}

/*****************************************************************************
** Internal: CopyFeedback()
** Short:    This routine is called during the decompression of files.
*/

static LONG CopyFeedback(OBJECTPTR Compression, struct CompressionFeedback *Feedback)
{
   UBYTE buffer[160];
   struct FileTime *date;
   struct acWrite write;
   struct Install *Self;
   OBJECTPTR file;
   LARGE srcstamp, deststamp;
   LONG len, result;

   result = CFR_Okay;
   if (Feedback->Progress IS 0) {
      if (AccessObject(GetOwner(Compression), 0, (OBJECTPTR *)&Self) IS ERR_Okay) {
         if (glForceCopy IS FALSE) {
            /* Check the file date stamps of the source and destination.  If the
            ** destination is more recent than the source, do not overwrite it.
            */

            if (AnalyseLocation(Feedback->Destination, NULL) IS ERR_Okay) {
               if (CreateObject(ID_FILE, NF_CHILD, &file, NULL,
                  FID_Location|TSTRING, Feedback->Destination,
                  TAGEND) IS ERR_Okay) {

                  if ((GetField(file, FID_Date, FT_POINTER, &date) IS ERR_Okay) AND (date)) {
                     /*** Is the destination more recent than the source? ***/

                     //DPrintF("Install:","Src: %d/%d/%d %d:%d:%d", Feedback->Day, Feedback->Month, Feedback->Year, Feedback->Hour, Feedback->Minute, Feedback->Second);
                     //DPrintF("Install:","Dst: %d/%d/%d %d:%d:%d", date->Day, date->Month, date->Year, date->Hour, date->Minute, date->Second);

                     srcstamp = ((LARGE)Feedback->Minute * 60) +
                              ((LARGE)Feedback->Hour * 60 * 60) +
                              ((LARGE)Feedback->Day * 60 * 60 * 24) +
                              ((LARGE)Feedback->Month * 60 * 60 * 24 * 31) +
                              ((LARGE)Feedback->Year * 60 * 60 * 24 * 31 * 12);

                     deststamp = ((LARGE)date->Minute * 60) +
                              ((LARGE)date->Hour * 60 * 60) +
                              ((LARGE)date->Day * 60 * 60 * 24) +
                              ((LARGE)date->Month * 60 * 60 * 24 * 31) +
                              ((LARGE)date->Year * 60 * 60 * 24 * 31 * 12);

                     if (deststamp > srcstamp) result = CFR_Skip;
                  }

                  Action(AC_Free, file, NULL);
               }
            }
         }

         /*** Send output to the user ***/

         if (result IS CFR_Skip) {
            len = StrFormat(buffer, sizeof(buffer), "Skipping file \"%s\" - destination more recent.", Feedback->Path, Feedback->Destination);
            ActionTags(AC_DataChannel, Self->prvOutput, Self->Head.UniqueID, DATA_TEXT, 1, buffer, len + 1, 1);
         }
         else {
            len = StrFormat(buffer, sizeof(buffer), "Decompressing \"%s\" to \"%s\".", Feedback->Path, Feedback->Destination);
            ActionTags(AC_DataChannel, Self->prvOutput, Self->Head.UniqueID, DATA_TEXT, 1, buffer, len + 1, 1);

            /*** Add the equivalent uninstall command ***/

            if (Self->prvUninstall) {
               for (len=0; Feedback->Destination[len]; len++);
               if ((Feedback->Destination[len-1] IS '/') OR (Feedback->Destination[len-1] IS '\\')) {
                  write.Length = StrFormat(buffer, sizeof(buffer), "  <delete src=\"%s\" empty/>\n", Feedback->Destination);
               }
               else write.Length = StrFormat(buffer, sizeof(buffer), "  <delete src=\"%s\"/>\n", Feedback->Destination);
               write.Buffer = buffer;
               Action(AC_Write, Self->prvUninstall, &write);
            }
         }

         ReleaseObject(Self);
      }
   }

   return(result);
}

/*****************************************************************************
** Internal: CommandDelete()
**
** <delete src="athene:reamde.txt"/>
*/

static ERROR CommandDelete(struct Install *Self, OBJECTPTR Output, struct XMLTag *XMLTag)
{
   UBYTE buffer[160];
   OBJECTPTR file;
   LONG i, len;

   for (i=1; i < XMLTag->TotalAttributes; i++) {
      if (StrCompare("src", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         if (CreateObject(ID_FILE, NF_CHILD, &file, NULL,
            FID_Location|TSTRING, XMLTag->Attributes[i].Value,
            TAGEND) IS ERR_Okay) {

            /*** Tell the user what we are doing ***/

            len = StrFormat(buffer, sizeof(buffer), "Deleting \"%s\".", XMLTag->Attributes[i].Value);
            ActionTags(AC_DataChannel, Output, Self->Head.UniqueID, DATA_TEXT, 1, buffer, len + 1, 1);

            /*** Delete the file ***/

            Action(MT_DeleteFile, file, NULL);

            Action(AC_Free, file, NULL);
         }
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: CommandDirectory()
*/

static ERROR CommandDirectory(struct Install *Self, OBJECTPTR Output, struct XMLTag *XMLTag)
{
   struct XMLTag *tag;
   OBJECTPTR info, input;
   OBJECTID browse_id, dialog_id, onclick_id;
   STRING content;
   LONG i, height;

   Self->prvDirectories[Self->prvDirectoryCount].Name = NULL;
   Self->prvDirectories[Self->prvDirectoryCount].Path = NULL;

   for (i=1; i < XMLTag->TotalAttributes; i++) {
      if (StrCompare("path", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         Self->prvDirectories[Self->prvDirectoryCount].Path = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("name", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         Self->prvDirectories[Self->prvDirectoryCount].Name = XMLTag->Attributes[i].Value;
      }
      else DPrintF("@Installer:","[Install] Unsupported directory attribute: %s.", XMLTag->Attributes[i].Name);
   }

   content = NULL;
   for (tag=XMLTag->Child; tag; tag=tag->Next) {
      if (!tag->Attributes[0].Name) { content = tag->Attributes[0].Value; break; }
   }

   if ((Self->prvDirectories[Self->prvDirectoryCount].Name) AND (Self->prvDirectories[Self->prvDirectoryCount].Path)) {

      if (content) {
         CreateObject(ID_TEXT, NULL, &info, NULL,
            FID_Owner|TLONG,     Self->ViewID,
            FID_String|TSTRING,  content,
            FID_XCoord|TLONG,    Self->prvCurrentXCoord,
            FID_YCoord|TLONG,    Self->prvCurrentYCoord,
            FID_Face|TSTRING,    glFontFace,
            FID_Colour|TSTRING,  "#000000",
            TAGEND);

         GetField(info, FID_TextHeight, FT_LONG, &height);
         Self->prvCurrentYCoord += height + 10;

         ActionMsg(AC_Draw, Self->ViewID, NULL);
      }

      CreateObject(ID_INPUT, NULL, &input, NULL,
         FID_Owner|TLONG,       Self->ViewID,
         FID_XCoord|TLONG,      Self->prvCurrentXCoord + 20,
         FID_XOffset|TLONG,     20,
         FID_YCoord|TLONG,      Self->prvCurrentYCoord,
         FID_Height|TLONG,      20,
         FID_String|TSTRING,    Self->prvDirectories[Self->prvDirectoryCount].Path,
         FID_Highlight|TSTRING, "#202020",
         FID_Shadow|TSTRING,    "#202020",
         FID_Colour|TSTRING,    "#f0f0f0",
         TAGEND);

      Self->prvDirectories[Self->prvDirectoryCount].Text = input;

      /*** Recalculate the render object height based on the height of our font ***/

      GetField(input, FID_Height, FT_LONG, &height);

      Action(AC_Show, input, NULL);

      /*** Create a directory browse button ***/

      CreateObject(ID_RENDER, NULL, NULL, &browse_id,
         FID_Owner|TLONG,   Self->ViewID,
         FID_XOffset|TLONG, 0,
         FID_YCoord|TLONG,  Self->prvCurrentYCoord,
         FID_Height|TLONG,  height,
         FID_Width|TLONG,   16,
         TAGEND);

      CreateObject(ID_IMAGE, NULL, NULL, NULL,
         FID_Owner|TLONG,   browse_id,
         FID_Src|TSTRING,   "icons:folders/folder(16)",
         FID_Align|TSTRING, "vertical",
         TAGEND);

      CreateObject(ID_ONCLICK, NULL, NULL, &onclick_id,
         FID_Owner|TLONG, browse_id,
         TAGEND);

      CreateObject(ID_FILEDIALOG, NULL, NULL, &dialog_id,
         FID_Owner|TLONG,      onclick_id,
         FID_Static|TLONG,     TRUE,
         FID_Title|TSTRING,    "Select a Destination Folder",
         FID_Location|TSTRING, Self->prvDirectories[Self->prvDirectoryCount].Path,
         FID_Flags|TLONG,      FDF_SELECTDIR|FDF_NOFILES,
         FID_OkayText|TSTRING, "Select Dir",
         TAGEND);

      CreateObject(ID_SET, NULL, NULL, NULL,
         FID_Owner|TLONG,      dialog_id,
         FID_Object|TLONG,     input->UniqueID,
         FID_String|TUNLISTED, "[owner.location]",
         TAGEND);

      ActionMsg(AC_Show, browse_id, NULL);

      Self->prvCurrentYCoord += height + 14;

      Self->prvDirectoryCount++;
   }
   else DPrintF("@Installer:","[Install] Error in <destination> tag, missing compulsory attributes.");

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: CommandItem()
*/

static ERROR CommandItem(struct Install *Self, OBJECTPTR Output, struct XMLTag *XMLTag)
{
   UBYTE buffer[200];
   struct acWrite write;
   struct XMLTag *tag;
   STRING content, item;
   LONG i;
   WORD method;

   if (!Self->prvConfig) return(ERR_Okay);

   item   = NULL;
   method = METHOD_OVERWRITE;
   for (i=1; i < XMLTag->TotalAttributes; i++) {
      if (StrCompare("name", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         item = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("method", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         method = TranslateMethod(XMLTag->Attributes[i].Value);
      }
      else DPrintF("@Installer:","[Install] Unsupported <item> attribute: %s.", XMLTag->Attributes[i].Name);
   }

   content = NULL;
   for (tag=XMLTag->Child; tag; tag=tag->Next) {
      if (!tag->Attributes[0].Name) { content = tag->Attributes[0].Value; break; }
   }

   if (Self->prvUninstall) {
      /* If the item already exists then record its value prior to us overwriting it.
      ** Otherwise, just tell the uninstaller to delete the item.
      */

      write.Buffer = buffer;
      write.Length = StrFormat(buffer, sizeof(buffer), "    <item name=\"%s\"/>\n", item);
      if (Action(AC_Write, Self->prvUninstall, &write) != ERR_Okay) return(ERR_Write);
   }

   ActionTags(MT_WriteConfig, Self->prvConfig, Self->prvMergeSection, item, content);

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: CommandMerge()
*/

static ERROR CommandMerge(struct Install *Self, OBJECTPTR Output, struct XMLTag *XMLTag)
{
   UBYTE buffer[300];
   struct XMLTag *tag;
   struct acWrite write;
   STRING file;
   LONG i, method, nofail, len;
   ERROR error;

   if (!XMLTag->Attributes) return(ERR_Okay);

   DPrintF("~Installer:","Processing <merge> tag.");

   file    = NULL;
   Self->prvMergeSection = NULL;
   method  = METHOD_OVERWRITE;
   nofail  = FALSE;

   for (i=1; i < XMLTag->TotalAttributes; i++) {
      if (StrCompare("file", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         file = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("section", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         Self->prvMergeSection = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("method", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         method = TranslateMethod(XMLTag->Attributes[i].Value);
      }
      else if (StrCompare("nofail", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         nofail = TRUE;
      }
   }

   error = ERR_Okay;
   if ((file) AND (Self->prvMergeSection)) {
      /*** Tell the user what we are doing ***/

      len = StrFormat(buffer, sizeof(buffer), "Merging section \"%s\" into configuration file \"%s\".", Self->prvMergeSection, file);
      ActionTags(AC_DataChannel, Output, Self->Head.UniqueID, DATA_TEXT, 1, buffer, len + 1, 1);

      if (Self->prvUninstall) {
         write.Length = StrFormat(buffer, sizeof(buffer), "  <merge file=\"%s\" section=\"%s\">\n", file, Self->prvMergeSection);
         write.Buffer = buffer;
         if ((error = Action(AC_Write, Self->prvUninstall, &write)) != ERR_Okay) goto exit;
      }

      /*** Open the configuration file ***/

      if ((error = CreateObject(ID_CONFIG, NF_CHILD, &Self->prvConfig, NULL,
            FID_Location|TSTRING, file,
            TAGEND)) IS ERR_Okay) {

         for (tag=XMLTag->Child; tag; tag=tag->Next) {
            for (i=0; glInstallCommands[i].Tag; i++) {
               if (StrCompare(glInstallCommands[i].Tag, tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                  if ((error = glInstallCommands[i].Routine(Self, Output, tag)) != ERR_Okay) goto exit;
                  break;
               }
            }
         }

         /*** Save configuration file changes ***/
         SaveObjectToFile(Self->prvConfig, file, PERMIT_EVERYONE_READWRITE);

         Action(AC_Free, Self->prvConfig, NULL);
         Self->prvConfig = NULL;
      }

      if (Self->prvUninstall) {
         write.Buffer = "  </merge>\n";
         for (i=0; write.Buffer[i]; i++);
         write.Length = i;
         if ((error = Action(AC_Write, Self->prvUninstall, &write)) != ERR_Okay) goto exit;
      }
   }
   else DPrintF("@Installer:","[Install] A <method> tag is missing file and/or section attributes.");

exit:
   StepBack();
   if (nofail) return(ERR_Okay);
   else return(error);
}

/*****************************************************************************
** Internal: CommandMove()
*/

static ERROR CommandMove(struct Install *Self, OBJECTPTR Output, struct XMLTag *XMLTag)
{
   UBYTE buffer[160];
   OBJECTPTR file;
   LONG i, len;
   STRING src, dest;

   src = NULL;
   dest = NULL;
   for (i=1; i < XMLTag->TotalAttributes; i++) {
      if (StrCompare("src", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         src = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("dest", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         dest = XMLTag->Attributes[i].Value;
      }
   }

   if ((src) AND (dest)) {
      if (CreateObject(ID_FILE, NF_CHILD, &file, NULL,
         FID_Location|TSTRING, src,
         TAGEND) IS ERR_Okay) {

         /*** Tell the user what we are doing ***/

         len = StrFormat(buffer, sizeof(buffer), "Moving \"%s\" to \"%s\".", src, dest);
         ActionTags(AC_DataChannel, Output, Self->Head.UniqueID, DATA_TEXT, 1, buffer, len + 1, 1);

         /*** Move the file ***/

         ActionTags(MT_MoveFile, file, dest);

         Action(AC_Free, file, NULL);
      }
   }
   else DPrintF("@Install:","<rename> tag is missing src and/or dest attributes.");

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: CommandOption()
*/

static ERROR CommandOption(struct Install *Self, OBJECTPTR Output, struct XMLTag *XMLTag)
{
   struct XMLTag *tag;
   STRING text;
   LONG i;
   ERROR error;

   Self->prvOptions[Self->prvOptionCount].Name = NULL;
   Self->prvOptions[Self->prvOptionCount].OptionStatus = FALSE;
   text = NULL;

   for (i=1; i < XMLTag->TotalAttributes; i++) {
      if (StrCompare("value", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         if (StrCompare("on", XMLTag->Attributes[i].Value, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            Self->prvOptions[Self->prvOptionCount].OptionStatus = TRUE;
         }
         else Self->prvOptions[Self->prvOptionCount].OptionStatus = FALSE;
      }
      else if (StrCompare("name", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         Self->prvOptions[Self->prvOptionCount].Name = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("text", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         text = XMLTag->Attributes[i].Value;
      }
      else DPrintF("@Installer:","[Install] Unsupported <option> attribute: %s.", XMLTag->Attributes[i].Name);
   }

   if ((Self->prvOptions[Self->prvOptionCount].Name) AND (text)) {
      Self->prvOptions[Self->prvOptionCount].Level = Self->prvIndentLevel;
      Self->prvOptions[Self->prvOptionCount].XCoord = Self->prvCurrentXCoord;
      Self->prvOptions[Self->prvOptionCount].YCoord = Self->prvCurrentYCoord;
      if (CreateOption(Self, &Self->prvOptions[Self->prvOptionCount].ButtonID, text, Self->prvOptions[Self->prvOptionCount].OptionStatus) IS ERR_Okay) {
         Self->prvOptionCount++;
      }
   }
   else DPrintF("@Installer:","[Install] Error in <option> tag, missing compulsory attributes.");

   Self->prvCurrentXCoord += Self->prvIndent;
   Self->prvIndentLevel++;

   for (tag=XMLTag->Child; tag; tag=tag->Next) {
      for (i=0; glInstallCommands[i].Tag; i++) {
         if (StrCompare(glInstallCommands[i].Tag, tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            if ((error = glInstallCommands[i].Routine(Self, Output, tag)) != ERR_Okay) return(error);
            break;
         }
      }
   }

   Self->prvIndentLevel--;
   Self->prvCurrentXCoord -= Self->prvIndent;

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: CommandRegister()
*/

static ERROR CommandRegister(struct Install *Self, OBJECTPTR Output, struct XMLTag *XMLTag)
{
   struct acWrite write;
   UBYTE buffer[300];
   OBJECTPTR config, userconfig;
   OBJECTID systemclasses;
   STRING classname, classid, parentid, category, dest, src, program, extension, header;
   LONG i, len, exists;
   ERROR error;

   if (!XMLTag->Attributes) return(ERR_Okay);

   DPrintF("~Installer:","Processing <register> tag.");

   classname = NULL;
   parentid  = NULL;
   classid   = NULL;
   category  = NULL;
   extension = NULL;
   header    = NULL;
   dest      = NULL;
   src       = NULL;
   program   = NULL;
   exists    = FALSE;

   for (i=1; i < XMLTag->TotalAttributes; i++) {
      if (StrCompare("class", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         classname = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("id", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         classid = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("parentid", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         parentid = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("extension", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         extension = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("header", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         header = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("category", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         category = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("dest", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         dest = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("src", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         src = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("program", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         program = XMLTag->Attributes[i].Value;
      }
   }

   error = ERR_Okay;
   if ((classname) AND (classid) AND (category) AND (dest) AND (src)) {
      /*** Tell the user what we are doing ***/

      len = StrFormat(buffer, sizeof(buffer), "Registering class \"%s\".", classname);
      ActionTags(AC_DataChannel, Output, Self->Head.UniqueID, DATA_TEXT, 1, buffer, len + 1, 1);

      if ((error = FastFindObject("SystemClasses", ID_CONFIG, &systemclasses, 1, NULL)) IS ERR_Okay) {
         if ((error = AccessObject(systemclasses, 5000, &config)) IS ERR_Okay) {
            if (parentid) ActionTags(MT_WriteConfig, config, classname, "ParentID", parentid);
            ActionTags(MT_WriteConfig, config, classname, "ClassID", classid);
            ActionTags(MT_WriteConfig, config, classname, "Category", category);
            ActionTags(MT_WriteConfig, config, classname, "Location", dest);
            if (extension) ActionTags(MT_WriteConfig, config, classname, "Extension", extension);
            if (header) ActionTags(MT_WriteConfig, config, classname, "Header", header);
            ReleaseObject(config);
         }
         else goto exit;

         if ((error = CreateObject(ID_CONFIG, NF_CHILD, &config, NULL,
            FID_Location|TSTRING, "system:config/kernel/classes.cfg",
            TAGEND)) IS ERR_Okay) {

            if (parentid) ActionTags(MT_WriteConfig, config, classname, "ParentID", parentid);
            ActionTags(MT_WriteConfig, config, classname, "ClassID", classid);
            ActionTags(MT_WriteConfig, config, classname, "Category", category);
            ActionTags(MT_WriteConfig, config, classname, "Location", dest);
            if (extension) ActionTags(MT_WriteConfig, config, classname, "Extension", extension);
            if (header) ActionTags(MT_WriteConfig, config, classname, "Header", header);

            SaveObjectToFile(config, "system:config/kernel/classes.cfg", PERMIT_EVERYONE_READWRITE);

            Action(AC_Free, config, NULL);
         }
         else goto exit;

         error = ActionTags(MT_DecompressFile, Self->prvCompression, src, dest, NULL);
      }
      else goto exit;

      if ((Self->prvUninstall) AND (exists IS FALSE)) {
         write.Length = StrFormat(buffer, sizeof(buffer), "  <unregister class=\"%s\" location=\"%s\"/>\n", classname, dest);
         write.Buffer = buffer;
         Action(AC_Write, Self->prvUninstall, &write);
      }
   }
   else if (program) {
      /*** Tell the user what we are doing ***/

      len = StrFormat(buffer, sizeof(buffer), "Registering program \"%s\".", program);
      ActionTags(AC_DataChannel, Output, Self->Head.UniqueID, DATA_TEXT, 1, buffer, len + 1, 1);

      /*** Register the program in the global program configuration file ***/

      if (CreateObject(ID_CONFIG, NF_CHILD, &config, NULL,
         FID_Location|TSTRING, "system:config/software/programs.cfg",
         TAGEND) IS ERR_Okay) {

         RegisterProgram(Self, XMLTag, config, program);

         if (SaveObjectToFile(config, "system:config/software/programs.cfg", PERMIT_EVERYONE_READWRITE) != ERR_Okay) {
            /* If the user does not have write access to the global programs.cfg
            ** file, we add the configuration information to his personal
            ** programs.cfg file instead.
            */

            if (CreateObject(ID_CONFIG, NF_CHILD, &userconfig, NULL,
               FID_Location|TSTRING, "user:config/programs.cfg",
               TAGEND) IS ERR_Okay) {

               RegisterProgram(Self, XMLTag, userconfig, program);

               SaveObjectToFile(userconfig, "user:config/programs.cfg", PERMIT_READ|PERMIT_WRITE);

               Action(AC_Free, userconfig, NULL);
            }
         }

         Action(AC_Free, config, NULL);
      }

      if (Self->prvUninstall) {
         write.Buffer = buffer;
         write.Length = StrFormat(buffer, sizeof(buffer), "  <unregister program=\"%s\"/>\n", program);
         Action(AC_Write, Self->prvUninstall, &write);
      }
   }
   else DPrintF("@Installer:","[Install] A <register> tag is missing required attributes.");

exit:
   StepBack();
   return(error);
}

static ERROR RegisterProgram(struct Install *Self, struct XMLTag *XMLTag, OBJECTPTR Config, STRING Section)
{
   struct XMLTag *tag;
   STRING content;

   for (tag=XMLTag->Child; tag; tag=tag->Next) {
      if (!tag->Child) continue;
      if (!(content = tag->Child->Attributes[0].Value)) continue;

      if (StrCompare("desktop", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         if ((StrCompare("1", content, 0, STR_MATCHLENGTH) IS ERR_Okay) OR (StrCompare("Yes", content, 0, STR_MATCHLENGTH) IS ERR_Okay)) {
            ActionTags(MT_WriteConfig, Config, Section, "Desktop", "Yes");
            Self->prvRefreshDesktop = TRUE;
         }
         else ActionTags(MT_WriteConfig, Config, Section, "Desktop", "No");
      }
      else if (StrCompare("text", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         ActionTags(MT_WriteConfig, Config, Section, "Text", content);
      }
      else if (StrCompare("description", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         ActionTags(MT_WriteConfig, Config, Section, "Description", content);
      }
      else if (StrCompare("icon", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         ActionTags(MT_WriteConfig, Config, Section, "Icon", content);
      }
      else if (StrCompare("command", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         ActionTags(MT_WriteConfig, Config, Section, "Command", content);
      }
      else if (StrCompare("category", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         ActionTags(MT_WriteConfig, Config, Section, "Category", content);
      }
      else DPrintF("@Installer:","[Install] Unrecognised <register> child tag \"%s\".", tag->Attributes[0].Name);
   }

   if (Self->prvPackageName) {
      ActionTags(MT_WriteConfig, Config, Section, "Package", Self->prvPackageName);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: CommandRename()
*/

static ERROR CommandRename(struct Install *Self, OBJECTPTR Output, struct XMLTag *XMLTag)
{
   UBYTE buffer[160];
   OBJECTPTR file;
   LONG i, len;
   STRING src, newname;

   src = NULL;
   newname = NULL;
   for (i=1; i < XMLTag->TotalAttributes; i++) {
      if (StrCompare("src", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         src = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("newname", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         newname = XMLTag->Attributes[i].Value;
      }
   }

   if ((src) AND (newname)) {
      if (CreateObject(ID_FILE, NF_CHILD, &file, NULL,
         FID_Location|TSTRING, src,
         TAGEND) IS ERR_Okay) {

         /*** Tell the user what we are doing ***/

         len = StrFormat(buffer, sizeof(buffer), "Renaming \"%s\" to \"%s\".", src, newname);
         ActionTags(AC_DataChannel, Output, Self->Head.UniqueID, DATA_TEXT, 1, buffer, len + 1, 1);

         /*** Rename the file ***/

         ActionTags(AC_Rename, file, newname);

         Action(AC_Free, file, NULL);
      }
   }
   else DPrintF("@Install:","<rename> tag is missing src and/or newname attributes.");

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: CommandRun()
*/

static ERROR CommandRun(struct Install *Self, OBJECTPTR Output, struct XMLTag *XMLTag)
{
   UBYTE buffer[160];
   OBJECTPTR run;
   STRING src, args;
   ERROR error;
   LONG i, len, nofail;

   src = NULL;
   args = "";
   nofail = FALSE;
   if (NewObject(ID_RUN, NF_CHILD, &run, NULL) IS ERR_Okay) {

      for (i=1; i < XMLTag->TotalAttributes; i++) {
         if (StrCompare("src", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            SetFieldVariable(run, "src", XMLTag->Attributes[i].Value);
            src = XMLTag->Attributes[i].Value;
         }
         else if (StrCompare("args",  XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            SetField(run, FID_Args, FT_STRING, XMLTag->Attributes[i].Value);
            args = XMLTag->Attributes[i].Value;
         }
         else if (StrCompare("mode", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            SetFieldVariable(run, "mode", XMLTag->Attributes[i].Value);
         }
         else if (StrCompare("flags", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            SetFieldVariable(run, "flags", XMLTag->Attributes[i].Value);
         }
         else if (StrCompare("nofail", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            nofail = TRUE;
         }
      }

      /*** Tell the user what we are doing ***/

      if (src) {
         len = StrFormat(buffer, sizeof(buffer), "%s %s", src, args);
         ActionTags(AC_DataChannel, Output, Self->Head.UniqueID, DATA_TEXT, 1, buffer, len + 1, 1);

         if ((error = Action(AC_Init, run, NULL)) IS ERR_Okay) {
            if ((error = Action(AC_Activate, run, NULL)) IS ERR_Okay) {
               Action(AC_Free, run, NULL);
               return(ERR_Okay);
            }
         }
      }
      else error = ERR_Okay;

      Action(AC_Free, run, NULL);
   }
   else error = ERR_NewObject;

   if (nofail IS TRUE) return(ERR_Okay);
   else return(error);
}

/*****************************************************************************
** Internal: CommandSection()
*/

static ERROR CommandSection(struct Install *Self, OBJECTPTR Output, struct XMLTag *XMLTag)
{
   struct acWrite write;
   UBYTE buffer[80];
   LONG i;

   for (i=1; i < XMLTag->TotalAttributes; i++) {
      if (StrCompare("name", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         Self->prvMergeSection = XMLTag->Attributes[i].Value;
      }
      else DPrintF("@Installer:","[Install] Unsupported <section> attribute: %s.", XMLTag->Attributes[i].Name);
   }

   if (Self->prvUninstall) {
      write.Length = StrFormat(buffer, sizeof(buffer), "  <section name=\"%s\"/>\n", Self->prvMergeSection);
      write.Buffer = buffer;
      return(Action(AC_Write, Self->prvUninstall, &write));
   }

   return(ERR_Okay);
}
